@@ -1,8 +1,8 @@
 ;;; ghub-graphql.el --- Access Github API using GrapthQL  -*- lexical-binding:t -*-
 
-;; Copyright (C) 2016-2023 Jonas Bernoulli
+;; Copyright (C) 2016-2024 Jonas Bernoulli
 
-;; Author: Jonas Bernoulli <jonas@bernoul.li>
+;; Author: Jonas Bernoulli <emacs.ghub@jonas.bernoulli.dev>
 ;; Homepage: https://github.com/magit/ghub
 ;; Keywords: tools
 
@@ -37,16 +37,19 @@
 
 ;;; Api
 
-(defvar ghub-graphql-items-per-request 100
+(define-error 'ghub-graphql-error "GraphQL Error" 'ghub-error)
+
+(defvar ghub-graphql-items-per-request 50
   "Number of GraphQL items to query for entities that return a collection.
 
 Adjust this value if you're hitting query timeouts against larger
 repositories.")
 
-(cl-defun ghub-graphql (graphql &optional variables
-                                &key username auth host forge
-                                headers silent
-                                callback errorback value extra)
+(cl-defun ghub-graphql (graphql
+                        &optional variables
+                        &key username auth host forge
+                        headers silent
+                        callback errorback value extra)
   "Make a GraphQL request using GRAPHQL and VARIABLES.
 Return the response as a JSON-like alist.  Even if the response
 contains `errors', do not raise an error.  GRAPHQL is a GraphQL
@@ -126,6 +129,8 @@ behave as for `ghub-request' (which see)."
                      number
                      id
                      state
+                     stateReason
+                     isReadByViewer
                      (author login)
                      title
                      createdAt
@@ -137,6 +142,7 @@ behave as for `ghub-request' (which see)."
                      (assignees [(:edges t)]
                                 id)
                      (comments  [(:edges t)]
+                                id
                                 databaseId
                                 (author login)
                                 createdAt
@@ -166,6 +172,7 @@ behave as for `ghub-request' (which see)."
                      number
                      id
                      state
+                     isReadByViewer
                      (author login)
                      title
                      createdAt
@@ -190,6 +197,7 @@ behave as for `ghub-request' (which see)."
                      (reviewRequests [(:edges t)]
                                      (requestedReviewer "... on User { id }\n"))
                      (comments  [(:edges t)]
+                                id
                                 databaseId
                                 (author login)
                                 createdAt
@@ -226,10 +234,10 @@ behave as for `ghub-request' (which see)."
                                               (originalCommit oid)
                                               path))))))
 
-(cl-defun ghub-fetch-repository (owner name callback
-                                       &optional until
-                                       &key username auth host forge
-                                       headers errorback sparse)
+(cl-defun ghub-fetch-repository ( owner name callback
+                                  &optional until
+                                  &key username auth host forge
+                                  headers paginate errorback sparse)
   "Asynchronously fetch forge data about the specified repository.
 Once all data has been collected, CALLBACK is called with the
 data as the only argument."
@@ -245,16 +253,17 @@ data as the only argument."
                         :host     host
                         :forge    forge
                         :headers  headers
+                        :paginate paginate
                         :errorback errorback))
 
-(cl-defun ghub-fetch-issue (owner name number callback
-                                  &optional until
-                                  &key username auth host forge
-                                  headers errorback)
+(cl-defun ghub-fetch-issue ( owner name number callback
+                             &optional until
+                             &key username auth host forge
+                             headers paginate errorback)
   "Asynchronously fetch forge data about the specified issue.
 Once all data has been collected, CALLBACK is called with the
 data as the only argument."
-  (ghub--graphql-vacuum (ghub--graphql-prepare-query
+  (ghub--graphql-vacuum (ghub--graphql-narrow-query
                          ghub-fetch-repository
                          `(repository issues (issue . ,number)))
                         `((owner . ,owner)
@@ -266,16 +275,17 @@ data as the only argument."
                         :host     host
                         :forge    forge
                         :headers  headers
+                        :paginate paginate
                         :errorback errorback))
 
-(cl-defun ghub-fetch-pullreq (owner name number callback
-                                    &optional until
-                                    &key username auth host forge
-                                    headers errorback)
+(cl-defun ghub-fetch-pullreq ( owner name number callback
+                               &optional until
+                               &key username auth host forge
+                               headers paginate errorback)
   "Asynchronously fetch forge data about the specified pull-request.
 Once all data has been collected, CALLBACK is called with the
 data as the only argument."
-  (ghub--graphql-vacuum (ghub--graphql-prepare-query
+  (ghub--graphql-vacuum (ghub--graphql-narrow-query
                          ghub-fetch-repository
                          `(repository pullRequests (pullRequest . ,number)))
                         `((owner . ,owner)
@@ -287,16 +297,17 @@ data as the only argument."
                         :host     host
                         :forge    forge
                         :headers  headers
+                        :paginate paginate
                         :errorback errorback))
 
-(cl-defun ghub-fetch-review-threads (owner name number callback
-                                           &optional until
-                                           &key username auth host forge
-                                           headers errorback)
+(cl-defun ghub-fetch-review-threads ( owner name number callback
+                                      &optional until
+                                      &key username auth host forge
+                                      headers paginate errorback)
   "Asynchronously fetch forge data about the review threads from a pull-request.
 Once all data has been collected, CALLBACK is called with the
 data as the only argument."
-  (ghub--graphql-vacuum (ghub--graphql-prepare-query
+  (ghub--graphql-vacuum (ghub--graphql-narrow-query
                          ghub-fetch-repository-review-threads
                          `(repository pullRequests (pullRequest . ,number)))
                         `((owner . ,owner)
@@ -308,10 +319,14 @@ data as the only argument."
                         :host     host
                         :forge    forge
                         :headers  headers
+                        :paginate paginate
                         :errorback errorback))
 
 ;;; Internal
 
+(defvar ghub--graphql-debug nil
+  "Whether `ghub--graphql-retrieve' updates the \" *gsexp-encode*\" buffer.")
+
 (cl-defstruct (ghub--graphql-req
                (:include ghub--req)
                (:constructor ghub--make-graphql-req)
@@ -320,13 +335,14 @@ data as the only argument."
   (query-str nil :read-only nil)
   (variables nil :read-only t)
   (until     nil :read-only t)
-  (buffer    nil :read-only t)
-  (pages     0   :read-only nil))
+  (pages     0   :read-only nil)
+  (paginate  nil :read-only nil)
+  (narrow    nil :read-only t))
 
-(cl-defun ghub--graphql-vacuum (query variables callback
-                                      &optional until
-                                      &key narrow username auth host forge
-                                      headers errorback)
+(cl-defun ghub--graphql-vacuum ( query variables callback
+                                 &optional until
+                                 &key narrow username auth host forge
+                                 headers paginate errorback)
   "Make a GraphQL request using QUERY and VARIABLES.
 See Info node `(ghub)GraphQL Support'."
   (unless host
@@ -347,22 +363,30 @@ See Info node `(ghub)GraphQL Support'."
     :variables variables
     :until     until
     :buffer    (current-buffer)
-    :callback  (and (not (eq callback 'synchronous))
-                    (let ((buf (current-buffer)))
-                      (if narrow
-                          (lambda (data)
-                            (let ((path narrow) key)
-                              (while (setq key (pop path))
-                                (setq data (cdr (assq key data)))))
-                            (ghub--graphql-set-mode-line buf nil)
-                            (funcall (or callback #'ghub--graphql-pp-response)
-                                     data))
-                        (lambda (data)
-                          (ghub--graphql-set-mode-line buf nil)
-                          (funcall (or callback #'ghub--graphql-pp-response)
-                                   data)))))
-    :errorback (and (not (eq callback 'synchronous))
-                    errorback))))
+    :narrow    narrow
+    :paginate  (or paginate
+                   (and-let* ((p (and (eq auth 'forge)
+                                      (fboundp 'magit-get)
+                                      (magit-get "forge.graphqlItemLimit"))))
+                     (string-to-number p)))
+    :callback  (and (not (eq callback 'synchronous)) callback)
+    :errorback (and (not (eq callback 'synchronous)) errorback))))
+
+(defvar ghub--graphql-synchronous-value nil)
+
+(cl-defun ghub--graphql-synchronous ( query variables
+                                      &optional until
+                                      &key narrow username auth host forge
+                                      headers paginate)
+  "Make a synchronous GraphQL request using QUERY and VARIABLES.
+See Info node `(ghub)GraphQL Support'."
+  (unwind-protect
+      (progn (ghub--graphql-vacuum query variables 'synchronous until
+                                   :narrow narrow :username username
+                                   :auth auth :host host :forge forge
+                                   :headers headers :paginate paginate)
+             ghub--graphql-synchronous-value)
+    (setq ghub--graphql-synchronous-value nil)))
 
 (cl-defun ghub--graphql-retrieve (req &optional lineage cursor)
   (let ((p (cl-incf (ghub--graphql-req-pages req))))
@@ -373,6 +397,10 @@ See Info node `(ghub)GraphQL Support'."
          (ghub--graphql-prepare-query
           (ghub--graphql-req-query req)
           lineage cursor)))
+  (when ghub--graphql-debug
+    (with-current-buffer (get-buffer-create " *gsexp-encode*")
+      (erase-buffer)
+      (insert (ghub--graphql-req-query-str req))))
   (ghub--retrieve
    (let ((json-false nil))
      (ghub--encode-payload
@@ -380,7 +408,7 @@ See Info node `(ghub)GraphQL Support'."
         (variables . ,(ghub--graphql-req-variables req)))))
    req))
 
-(defun ghub--graphql-prepare-query (query &optional lineage cursor)
+(defun ghub--graphql-prepare-query (query &optional lineage cursor paginate)
   (when lineage
     (setq query (ghub--graphql-narrow-query query lineage cursor)))
   (let ((loc (ghub--alist-zip query))
@@ -392,8 +420,14 @@ See Info node `(ghub)GraphQL Support'."
                      (listp (aref node 0)))
             (let ((alist (cl-coerce node 'list))
                   vars)
-              (when (cadr (assq :edges alist))
-                (push (list 'first ghub-graphql-items-per-request) vars)
+              (when-let ((edges (cadr (assq :edges alist))))
+                (push (list 'first
+                            (apply
+                             #'min
+                             (delq nil (list (and (numberp edges) edges)
+                                             paginate
+                                             ghub-graphql-items-per-request))))
+                      vars)
                 (setq loc  (treepy-up loc))
                 (setq node (treepy-node loc))
                 (setq loc  (treepy-replace
@@ -419,24 +453,46 @@ See Info node `(ghub)GraphQL Support'."
               (cl-return node))
           (setq loc (treepy-next loc)))))))
 
-(defun ghub--graphql-handle-response (status req)
-  (let ((buffer (current-buffer)))
+(cl-defun ghub--graphql-handle-response (status req &optional (buffer nil sbuf))
+  (let ((buf (if sbuf
+                 (and (buffer-live-p buffer) buffer)
+               (and status (current-buffer)))))
     (unwind-protect
-        (progn
-          (set-buffer-multibyte t)
-          (let* ((headers (ghub--handle-response-headers status req))
-                 (payload (ghub--handle-response-payload req))
+        (save-current-buffer
+          (when buf
+            (set-buffer buf)
+            (set-buffer-multibyte t))
+          (let* ((headers (and buf (ghub--handle-response-headers status req)))
+                 (payload (and buf (ghub--handle-response-payload req)))
                  (payload (ghub--handle-response-error status payload req))
                  (err     (plist-get status :error))
                  (errors  (cdr (assq 'errors payload)))
                  (errors  (and errors (cons 'ghub-graphql-error errors))))
             (if (or err errors)
-                (if-let ((errorback (ghub--req-errorback req)))
-                    (funcall errorback (or err errors) headers status req)
-                  (ghub--signal-error (or err errors)))
+                (ghub--graphql-handle-failure
+                 req (or err errors) headers status)
               (ghub--graphql-walk-response req (assq 'data payload)))))
-      (when (buffer-live-p buffer)
-        (kill-buffer buffer)))))
+      (when (buffer-live-p buf)
+        (kill-buffer buf)))))
+
+(defun ghub--graphql-handle-failure (req errors headers status)
+  (if-let ((errorback (ghub--req-errorback req)))
+      (let ((buffer (ghub--req-buffer req)))
+        (with-current-buffer
+            (if (buffer-live-p buffer) buffer (current-buffer))
+          (funcall errorback errors headers status req)))
+    (ghub--signal-error errors)))
+
+(defun ghub--graphql-handle-success (req data)
+  (let ((callback (ghub--req-callback req))
+        (buffer   (ghub--req-buffer req))
+        (narrow   (ghub--graphql-req-narrow req)))
+    (while-let ((key (pop narrow)))
+      (setq data (cdr (assq key data))))
+    (with-current-buffer
+        (if (buffer-live-p buffer) buffer (current-buffer))
+      (funcall (or callback #'ghub--graphql-pp-response)
+               data))))
 
 (defun ghub--graphql-walk-response (req data)
   (let* ((loc (ghub--req-value req))
@@ -457,32 +513,33 @@ See Info node `(ghub)GraphQL Support'."
             (let-alist val
               (let* ((cursor (and .pageInfo.hasNextPage
                                   .pageInfo.endCursor))
-                     (until (cdr (assq (intern (format "%s-until" key))
-                                       (ghub--graphql-req-until req))))
-                     (nodes (mapcar #'cdar .edges))
-                     (nodes (if until
-                                (seq-take-while
-                                 (lambda (node)
-                                   (or (string> (cdr (assq 'updatedAt node))
-                                                until)
-                                       (setq cursor nil)))
-                                 nodes)
-                              nodes)))
-                (if cursor
-                    (progn
-                      (setf (ghub--req-value req) loc)
-                      (ghub--graphql-retrieve req
-                                              (ghub--graphql-lineage loc)
-                                              cursor)
-                      (cl-return))
-                  (setq loc (treepy-replace loc (cons key nodes))))))))
+                     (until  (cdr (assq (intern (format "%s-until" key))
+                                        (ghub--graphql-req-until req))))
+                     (nodes  (mapcar #'cdar .edges))
+                     (nodes  (if until
+                                 (seq-take-while
+                                  (lambda (node)
+                                    (or (string> (cdr (assq 'updatedAt node))
+                                                 until)
+                                        (setq cursor nil)))
+                                  nodes)
+                               nodes)))
+                (cond (cursor
+                       (setf (ghub--req-value req) loc)
+                       (ghub--graphql-retrieve req
+                                               (ghub--graphql-lineage loc)
+                                               cursor)
+                       (cl-return))
+                      ((setq loc (treepy-replace loc (cons key nodes)))))))))
         (cond ((not (treepy-end-p loc))
                (setq loc (treepy-next loc)))
               ((ghub--req-callback req)
-               (funcall (ghub--req-callback req)
-                        (treepy-root loc))
+               (ghub--graphql-handle-success req (treepy-root loc))
+               (ghub--graphql-set-mode-line req nil)
                (cl-return))
-              ((cl-return (treepy-root loc))))))))
+              (t
+               (setq ghub--graphql-synchronous-value (treepy-root loc))
+               (cl-return)))))))
 
 (defun ghub--graphql-lineage (loc)
   (let (lineage)
@@ -492,15 +549,14 @@ See Info node `(ghub)GraphQL Support'."
     lineage))
 
 (defun ghub--graphql-narrow-data (data lineage)
-  (let (key)
-    (while (setq key (pop lineage))
-      (if (consp (car lineage))
-          (progn (pop lineage)
-                 (setf data (cadr data)))
-        (setq data (assq key (cdr data))))))
+  (while-let ((key (pop lineage)))
+    (if (consp (car lineage))
+        (progn (pop lineage)
+               (setf data (cadr data)))
+      (setq data (assq key (cdr data)))))
   data)
 
-(defun ghub--graphql-narrow-query (query lineage cursor)
+(defun ghub--graphql-narrow-query (query lineage &optional cursor)
   (if (consp (car lineage))
       (let* ((child  (cddr query))
              (alist  (cl-coerce (cadr query) 'list))
@@ -546,14 +602,13 @@ See Info node `(ghub)GraphQL Support'."
         (make-node (lambda (_ children) children)))
     (treepy-zipper branchp #'identity make-node root)))
 
-(defun ghub--graphql-set-mode-line (buf string &rest args)
-  (when (ghub--graphql-req-p buf)
-    (setq buf (ghub--graphql-req-buffer buf)))
-  (when (buffer-live-p buf)
-    (with-current-buffer buf
-      (setq mode-line-process
-            (and string (concat " " (apply #'format string args))))
-      (force-mode-line-update t))))
+(defun ghub--graphql-set-mode-line (req string &rest args)
+  (let ((buffer (ghub--graphql-req-buffer req)))
+    (when (buffer-live-p buffer)
+      (with-current-buffer buffer
+        (setq mode-line-process
+              (and string (concat " " (apply #'format string args))))
+        (force-mode-line-update t)))))
 
 (defun ghub--graphql-pp-response (data)
   (require 'pp) ; needed for Emacs < 29.
