@@ -1,11 +1,18 @@
 ;;; ghub.el --- Client libraries for Git forge APIs  -*- lexical-binding:t -*-
 
-;; Copyright (C) 2016-2023 Jonas Bernoulli
+;; Copyright (C) 2016-2024 Jonas Bernoulli
 
-;; Author: Jonas Bernoulli <jonas@bernoul.li>
+;; Author: Jonas Bernoulli <emacs.ghub@jonas.bernoulli.dev>
 ;; Homepage: https://github.com/magit/ghub
 ;; Keywords: tools
 
+;; Package-Version: 4.1.1
+;; Package-Requires: (
+;;     (emacs "26.3")
+;;     (compat "30.0.0.0")
+;;     (let-alist "1.0.6")
+;;     (treepy "0.1.2"))
+
 ;; SPDX-License-Identifier: GPL-3.0-or-later
 
 ;; This file is free software: you can redistribute it and/or modify
@@ -119,6 +126,7 @@ See https://github.com/magit/ghub/pull/149.")
   (unpaginate nil :read-only nil)
   (noerror    nil :read-only t)
   (reader     nil :read-only t)
+  (buffer     nil :read-only t)
   (callback   nil :read-only t)
   (errorback  nil :read-only t)
   (value      nil :read-only nil)
@@ -130,17 +138,19 @@ See https://github.com/magit/ghub/pull/149.")
 
 (define-error 'ghub-error "Ghub/Url Error" 'error)
 (define-error 'ghub-http-error "HTTP Error" 'ghub-error)
+(define-error 'ghub-http-timeout "HTTP Timeout" 'ghub-http-error)
 
 (defvar ghub-response-headers nil
   "The headers returned in response to the last request.
 `ghub-request' returns the response body and stores the
 response headers in this variable.")
 
-(cl-defun ghub-head (resource &optional params
-                              &key query payload headers
-                              silent unpaginate noerror reader
-                              username auth host
-                              callback errorback extra)
+(cl-defun ghub-head (resource
+                     &optional params
+                     &key query payload headers
+                     silent unpaginate noerror reader
+                     username auth host
+                     callback errorback extra)
   "Make a `HEAD' request for RESOURCE, with optional query PARAMS.
 Like calling `ghub-request' (which see) with \"HEAD\" as METHOD."
   (ghub-request "HEAD" resource params
@@ -150,11 +160,12 @@ Like calling `ghub-request' (which see) with \"HEAD\" as METHOD."
                 :username username :auth auth :host host
                 :callback callback :errorback errorback :extra extra))
 
-(cl-defun ghub-get (resource &optional params
-                             &key query payload headers
-                             silent unpaginate noerror reader
-                             username auth host
-                             callback errorback extra)
+(cl-defun ghub-get (resource
+                    &optional params
+                    &key query payload headers
+                    silent unpaginate noerror reader
+                    username auth host
+                    callback errorback extra)
   "Make a `GET' request for RESOURCE, with optional query PARAMS.
 Like calling `ghub-request' (which see) with \"GET\" as METHOD."
   (ghub-request "GET" resource params
@@ -164,11 +175,12 @@ Like calling `ghub-request' (which see) with \"GET\" as METHOD."
                 :username username :auth auth :host host
                 :callback callback :errorback errorback :extra extra))
 
-(cl-defun ghub-put (resource &optional params
-                             &key query payload headers
-                             silent unpaginate noerror reader
-                             username auth host
-                             callback errorback extra)
+(cl-defun ghub-put (resource
+                    &optional params
+                    &key query payload headers
+                    silent unpaginate noerror reader
+                    username auth host
+                    callback errorback extra)
   "Make a `PUT' request for RESOURCE, with optional payload PARAMS.
 Like calling `ghub-request' (which see) with \"PUT\" as METHOD."
   (ghub-request "PUT" resource params
@@ -178,11 +190,12 @@ Like calling `ghub-request' (which see) with \"PUT\" as METHOD."
                 :username username :auth auth :host host
                 :callback callback :errorback errorback :extra extra))
 
-(cl-defun ghub-post (resource &optional params
-                              &key query payload headers
-                              silent unpaginate noerror reader
-                              username auth host
-                              callback errorback extra)
+(cl-defun ghub-post (resource
+                     &optional params
+                     &key query payload headers
+                     silent unpaginate noerror reader
+                     username auth host
+                     callback errorback extra)
   "Make a `POST' request for RESOURCE, with optional payload PARAMS.
 Like calling `ghub-request' (which see) with \"POST\" as METHOD."
   (ghub-request "POST" resource params
@@ -192,11 +205,12 @@ Like calling `ghub-request' (which see) with \"POST\" as METHOD."
                 :username username :auth auth :host host
                 :callback callback :errorback errorback :extra extra))
 
-(cl-defun ghub-patch (resource &optional params
-                               &key query payload headers
-                               silent unpaginate noerror reader
-                               username auth host
-                               callback errorback extra)
+(cl-defun ghub-patch (resource
+                      &optional params
+                      &key query payload headers
+                      silent unpaginate noerror reader
+                      username auth host
+                      callback errorback extra)
   "Make a `PATCH' request for RESOURCE, with optional payload PARAMS.
 Like calling `ghub-request' (which see) with \"PATCH\" as METHOD."
   (ghub-request "PATCH" resource params
@@ -206,11 +220,12 @@ Like calling `ghub-request' (which see) with \"PATCH\" as METHOD."
                 :username username :auth auth :host host
                 :callback callback :errorback errorback :extra extra))
 
-(cl-defun ghub-delete (resource &optional params
-                                &key query payload headers
-                                silent unpaginate noerror reader
-                                username auth host
-                                callback errorback extra)
+(cl-defun ghub-delete (resource
+                       &optional params
+                       &key query payload headers
+                       silent unpaginate noerror reader
+                       username auth host
+                       callback errorback extra)
   "Make a `DELETE' request for RESOURCE, with optional payload PARAMS.
 Like calling `ghub-request' (which see) with \"DELETE\" as METHOD."
   (ghub-request "DELETE" resource params
@@ -220,11 +235,12 @@ Like calling `ghub-request' (which see) with \"DELETE\" as METHOD."
                 :username username :auth auth :host host
                 :callback callback :errorback errorback :extra extra))
 
-(cl-defun ghub-request (method resource &optional params
-                               &key query payload headers
-                               silent unpaginate noerror reader
-                               username auth host forge
-                               callback errorback value extra)
+(cl-defun ghub-request ( method resource
+                         &optional params
+                         &key query payload headers
+                         silent unpaginate noerror reader
+                         username auth host forge
+                         callback errorback value extra)
   "Make a request for RESOURCE and return the response body.
 
 Also place the response headers in `ghub-response-headers'.
@@ -362,6 +378,7 @@ Both callbacks are called with four arguments.
     :unpaginate unpaginate
     :noerror    noerror
     :reader     reader
+    :buffer     (current-buffer)
     :callback   callback
     :errorback  errorback
     :value      value
@@ -384,8 +401,9 @@ called after the next request has finished.  Use the function
   (and (assq 'next (ghub-response-link-relations req))
        (or (ghub--retrieve nil req) t)))
 
-(cl-defun ghub-wait (resource &optional duration
-                              &key username auth host forge)
+(cl-defun ghub-wait (resource
+                     &optional duration
+                     &key username auth host forge)
   "Busy-wait up to DURATION seconds for RESOURCE to become available.
 
 DURATION specifies how many seconds to wait at most.  It defaults
@@ -439,7 +457,9 @@ this function is called with nil for PAYLOAD."
               (setq-default ghub-response-headers headers))
             page)
         (cdr (assq 'link-alist ghub-response-headers)))
-    (and-let* ((rels (cdr (assoc "Link" (or headers ghub-response-headers)))))
+    (and-let* ((headers (or headers ghub-response-headers))
+               (rels (cdr (or (assoc "Link" headers)
+                              (assoc "link" headers)))))
       (mapcar (lambda (elt)
                 (pcase-let ((`(,url ,rel) (split-string elt "; ")))
                   (cons (intern (substring rel 5 -1))
@@ -469,29 +489,15 @@ Signal an error if the id cannot be determined."
 
 ;;;; Internal
 
-(defvar ghub-use-workaround-for-emacs-bug t
-  "Whether to work around Emacs bug debbugs#34341.
-
-First see https://github.com/magit/ghub/wiki/Known-Issues,
-for information about this bug and another related bug.
-
-Because our under of these bugs evolved over time, the possible
-values of this variable are a bit odd: If t, enable workaround if
-necessary (i.e., if Emacs < 26.3 and GnuTLS >= 3.6.3 are used).
-If `force', enable workaround even if that is believed to be
-unnecessary.  If nil, do not enable the workaround.  The default
-is t.")
-
 (defvar ghub-use-workaround-for-emacs-bug-54989 t
   "Whether to work around Emacs bug debbugs#54989.
 
-First see https://github.com/magit/ghub/wiki/Known-Issues,
-for information about this bug and another related bug.
+If t (the default), work around the bug if necessary (i.e., if not
+using Emacs 29 or later).  If nil, then don't work around the bug.
+Setting this variable only has an effect if it is done before
+`ghub' is loaded.
 
-If t, work around the bug if necessary (i.e., if not using Emacs'
-\"master\" branch).  If nil, then don't work around the bug.  The
-default is t.  Setting this variable only has an effect if it is
-done before `ghub' is loaded.")
+Also see https://github.com/magit/ghub/wiki/Known-Issues.")
 
 (cl-defun ghub--retrieve (payload req)
   (let ((url-request-extra-headers
@@ -502,37 +508,32 @@ done before `ghub' is loaded.")
         (url-show-status nil)
         (url     (ghub--req-url req))
         (handler (ghub--req-handler req))
-        (silent  (ghub--req-silent req))
-        (gnutls-algorithm-priority
-         (if (and ghub-use-workaround-for-emacs-bug
-                  (or (eq ghub-use-workaround-for-emacs-bug 'force)
-                      (and (not gnutls-algorithm-priority)
-                           (>= libgnutls-version 30603)
-                           (version< emacs-version "26.3")
-                           ;; (memq (ghub--req-forge req) '(github nil))
-                           )))
-             "NORMAL:-VERS-TLS1.3"
-           gnutls-algorithm-priority)))
+        (silent  (ghub--req-silent req)))
     (if (or (ghub--req-callback  req)
             (ghub--req-errorback req))
         (url-retrieve url handler (list req) silent)
-      (with-current-buffer
-          (url-retrieve-synchronously url silent)
-        (funcall handler (car url-callback-arguments) req)))))
+      (let ((buf (url-retrieve-synchronously url silent)))
+        (funcall handler
+                 (and buf (buffer-local-value 'url-callback-arguments buf))
+                 req buf)))))
 
-(defun ghub--handle-response (status req)
-  (let ((buffer (current-buffer)))
+(cl-defun ghub--handle-response (status req &optional (buffer nil sbuf))
+  (let ((buf (if sbuf
+                 (and (buffer-live-p buffer) buffer)
+               (and status (current-buffer)))))
     (unwind-protect
-        (progn
-          (set-buffer-multibyte t)
+        (save-current-buffer
+          (when buf
+            (set-buffer buf)
+            (set-buffer-multibyte t))
           (let* ((unpaginate (ghub--req-unpaginate req))
-                 (headers    (ghub--handle-response-headers status req))
-                 (payload    (ghub--handle-response-payload req))
-                 (payload    (ghub--handle-response-error status payload req))
-                 (value      (ghub--handle-response-value payload req))
-                 (prev       (ghub--req-url req))
-                 (next       (cdr (assq 'next (ghub-response-link-relations
-                                               req headers payload)))))
+                 (headers (and buf (ghub--handle-response-headers status req)))
+                 (payload (and buf (ghub--handle-response-payload req)))
+                 (payload (ghub--handle-response-error status payload req))
+                 (value   (ghub--handle-response-value payload req))
+                 (prev    (ghub--req-url req))
+                 (next    (cdr (assq 'next (ghub-response-link-relations
+                                            req headers payload)))))
             (when (numberp unpaginate)
               (cl-decf unpaginate))
             (setf (ghub--req-url req)
@@ -543,7 +544,8 @@ done before `ghub' is loaded.")
                      (or (eq unpaginate t)
                          (>  unpaginate 0))
                      (ghub-continue req))
-                (let ((callback  (ghub--req-callback req))
+                (let ((req-buf   (ghub--req-buffer req))
+                      (callback  (ghub--req-callback req))
                       (errorback (ghub--req-errorback req))
                       (err       (plist-get status :error)))
                   (cond ((and err errorback)
@@ -553,43 +555,48 @@ done before `ghub' is loaded.")
                                     errorback)
                                   err headers status req))
                         (callback
-                         (funcall callback value headers status req))
+                         (save-current-buffer
+                           (when (buffer-live-p req-buf)
+                             (set-buffer req-buf))
+                           (funcall callback value headers status req)))
                         (t value))))))
-      (when (buffer-live-p buffer)
-        (kill-buffer buffer)))))
+      (when (buffer-live-p buf)
+        (kill-buffer buf)))))
 
-(defun ghub--handle-response-headers (_status req)
-  (goto-char (point-min))
-  (forward-line 1)
-  (let (headers)
-    (when (memq url-http-end-of-headers '(nil 0))
-      (setq url-debug t)
-      (error "BUG: missing headers; but there's a patch for that \
-see https://github.com/magit/ghub/wiki/Known-Issues"))
-    (while (re-search-forward "^\\([^:]*\\): \\(.+\\)"
-                              url-http-end-of-headers t)
-      (push (cons (match-string 1)
-                  (match-string 2))
-            headers))
-    (setq headers (nreverse headers))
-    (goto-char (1+ url-http-end-of-headers))
-    (if (and req (or (ghub--req-callback req)
-                     (ghub--req-errorback req)))
-        (setq-local ghub-response-headers headers)
-      (setq-default ghub-response-headers headers))
-    headers))
+(defun ghub--handle-response-headers (status req)
+  (and status ; Request did not time out.
+       (let ((headers ()))
+         (when (memq url-http-end-of-headers '(nil 0))
+           (error "Ghub: BUG: No headers in response buffer: %S"
+                  (current-buffer)))
+         (goto-char (point-min))
+         (forward-line 1)
+         (while (re-search-forward "^\\([^:]*\\): \\(.+\\)"
+                                   url-http-end-of-headers t)
+           (push (cons (match-string 1)
+                       (match-string 2))
+                 headers))
+         (setq headers (nreverse headers))
+         (goto-char (1+ url-http-end-of-headers))
+         (if (and req (or (ghub--req-callback req)
+                          (ghub--req-errorback req)))
+             (setq-local ghub-response-headers headers)
+           (setq-default ghub-response-headers headers))
+         headers)))
 
 (defun ghub--handle-response-error (status payload req)
-  (let ((noerror (ghub--req-noerror req))
-        (err (plist-get status :error)))
-    (if err
-        (if noerror
-            (if (eq noerror 'return)
-                payload
-              (setcdr (last err) (list payload))
-              nil)
-          (ghub--signal-error err payload req))
-      payload)))
+  (if-let ((err (or (plist-get status :error)
+                    (and (not status) 'timeout))))
+      (if-let ((noerror (ghub--req-noerror req)))
+          (cond ((eq noerror 'return) payload)
+                ((eq err 'timeout) nil)
+                (t (setcdr (last err) (list payload))
+                   nil))
+        (if (eq err 'timeout)
+            (signal 'ghub-http-timeout
+                    (list nil nil (url-recreate-url (ghub--req-url req)) nil))
+          (ghub--signal-error err payload req)))
+    payload))
 
 (defun ghub--signal-error (err &optional payload req)
   (pcase-let ((`(,symb . ,data) err))
@@ -599,7 +606,7 @@ see https://github.com/magit/ghub/wiki/Known-Issues"))
                     (let ((code (car (cdr-safe data))))
                       (list code
                             (nth 2 (assq code url-http-codes))
-                            (and req (url-filename (ghub--req-url req)))
+                            (and req (url-recreate-url (ghub--req-url req)))
                             payload)))
           (signal 'ghub-error data))
       (signal symb data))))
@@ -822,11 +829,11 @@ or (info \"(ghub)Getting Started\") for instructions.
               (format "bitbucket.%s.user" host)))
            (gitea
             (when (zerop (call-process "git" nil nil nil "config" "gitea.host"))
-              (error "gitea.host is set but always ignored"))
+              (message "WARNING: gitea.host is set but always ignored"))
             (format "gitea.%s.user" host))
            (gogs
             (when (zerop (call-process "git" nil nil nil "config" "gogs.host"))
-              (error "gogs.host is set but always ignored"))
+              (message "WARNING: gogs.host is set but always ignored"))
             (format "gogs.%s.user"  host)))))
     (condition-case nil
         (car (process-lines "git" "config" var))
@@ -858,17 +865,6 @@ or (info \"(ghub)Getting Started\") for instructions.
               (plist-get plist k))
             keys)))
 
-(when (version< emacs-version "26.2")
-  ;; Fixed by Emacs commit 60ff8101449eea3a5ca4961299501efd83d011bd.
-  (advice-add 'auth-source-netrc-parse-next-interesting :around
-              'auth-source-netrc-parse-next-interesting@save-match-data)
-  (defun auth-source-netrc-parse-next-interesting@save-match-data (fn)
-    "Save match-data for the benefit of caller `auth-source-netrc-parse-one'.
-Without wrapping this function in `save-match-data' the caller
-won't see the secret from a line that is followed by a commented
-line."
-    (save-match-data (funcall fn))))
-
 (when (< emacs-major-version 28)
   ;; Fixed by Emacs commit 0b98ea5fbe276c67206896dca111c000f984ee0f.
   (advice-add 'url-http-handle-authentication :around
@@ -878,7 +874,7 @@ line."
 For Emacs 27.1 prevent a useful `http' error from being replaced
 by a generic one that omits all useful information.  For earlier
 releases prevent a new request from being made, which would
-either result in an infinite loop or (e.g. in the case of `ghub')
+either result in an infinite loop or (e.g., in the case of `ghub')
 the user being asked for their name."
     (if (assoc "Authorization" url-http-extra-headers)
         t ; Return "success", here also known as "successfully failed".
@@ -887,14 +883,18 @@ the user being asked for their name."
 (when (and (< emacs-major-version 29)
            ghub-use-workaround-for-emacs-bug-54989)
   ;; Fixed in Emacs commit 0829c6836eff14dda0cf8b3047376967f7b000f4.
+  ;; Discussed in https://debbugs.gnu.org/cgi/bugreport.cgi?bug=54989
+  ;; and https://github.com/magit/ghub/issues/81#issuecomment-1100700165.
+  ;;
   ;; Cleanup from 26faa2b943675107e1664b2fea7174137c473475 is not
   ;; included in this copy because doing that would require changes
-  ;; to more functions.  This function has seen a few other changes
-  ;; since Emacs 25.1, the oldest version we still support.  Of these
-  ;; only 4f1df40db36b221e7842bd75d6281922dcb268ee makes a functional
-  ;; change, fixing debbug#35658.  The first release to contain that
-  ;; commit is 27.1.  That commit either fixes a related bug or it
-  ;; deals with the same bug but only partially fixes it.
+  ;; to more functions.
+  ;;
+  ;; This function has seen a few other changes since Emacs 25.1.
+  ;; Of these only 4f1df40db36b221e7842bd75d6281922dcb268ee makes
+  ;; a functional change, fixing debbug#35658.  The first release to
+  ;; contain that commit is 27.1.  That commit either fixes a related
+  ;; bug or it deals with the same bug but only partially fixes it.
   (advice-add 'url-http-chunked-encoding-after-change-function :override
               'url-http-chunked-encoding-after-change-function@54989-backport)
   (defvar url-http-chunked-last-crlf-missing nil)
